#!/bin/bash
#
# Skript zur Installation des ULS (Universal Logging System)
#
#
# Schalter:
#   --force    : bestehende DB überschreiben
#   -C <file>  : mysql-defaults-file mit Informationen zur Anmeldung
#   -M         : MySQL kann ohne Kennwort aufgerufen werden
#   -D         : Defaults für -g -u und -w nutzen
#   -g <gid>   : Groupid der Gruppe 'ulsuser'
#   -u <uid>   : Userid des Users 'uls'
#   -w <uid>   : Userid des Users 'wwwuls'
#   -Y         : Ohne Rückfragen alles ausführen
############################################################
umask 022
#
ULSBASE="/srv"
unset mysqldefaults force nodbpwd dbpwd defaultids ulsgid ulsuid wwwulsuid allwaysyes
#
while getopts C:MDg:u:w:Y?-: op; do
  case "$op" in
    C) mysqldefaults="$OPTARG";;
    M) nodbpwd=1;;
    D) defaultids=1;;
    g) ulsgid="$OPTARG";;
    u) ulsuid="$OPTARG";;
    w) wwwulsuid="$OPTARG";;
    Y) allwaysyes=1;;
    -*) if [[ "$OPTARG" = "force" ]]
         then
          force=--force
        fi
       ;;
    *) echo "usage: `basename $0` [--force] [-C <mysql-defaults-file>] [-M] [-D] [-g <gid>] [-u <uid>] [-w <uid>] [-Y] [<uls-basedir>]"
       echo "  --force    : bestehende DB überschreiben
  -C <file>  : mysql-defaults-file
  -M         : MySQL kann ohne Kennwort aufgerufen werden
  -D         : Defaults für -g -u und -w nutzen
  -g <gid>   : Groupid der Gruppe 'ulsuser'
  -u <uid>   : Userid des Users 'uls'
  -w <uid>   : Userid des Users 'wwwuls'
  -Y         : Ohne Rückfragen alles ausführen"
       exit 1;;
  esac
done
shift $(( $OPTIND -1))
#
if [[ -n "$1" ]]; then
  ULSBASE="$1"
fi
export ULSBASE
ULSDIR="$ULSBASE/uls"
#
THISDIR=`dirname $0`
#
$THISDIR/inst_check.sh || exit
#
echo
echo "Es sollten gnuplot und wkhtmltopdf installiert sein."
echo
echo "Vor der Installation muss die /etc/[mysql/]my.cnf angepasst werden"
echo "Die Zeilen aus der my.cnf sollten entsprechend in der /etc/[mysql/]my.cnf"
echo "eingefuegt oder geaendert werden."
echo "Wichtig sind die max_allowed_packet und sql_mode Zeilen."
echo
#
if [[ -z "$allwaysyes" ]]; then
  echo "Wenn die Einstellungen nicht gemacht sind oder MySQL nach den"
  echo "Aenderungen noch nicht neu gestartet wurde, dann bitte jetzt"
  echo -n "die Installation abbrechen, sonst Return druecken. "
  read dummy
echo
fi
#
if [[ -z "$mysqldefaults" ]]; then
  if [[ -z "$nodbpwd" ]]; then
    echo -n "Bitte MySQL-Kennwort fuer User root eingeben (Return, wenn keins benötigt wird): "
    read -s dbpwd
    echo
  fi
  mysqlinstdefaults=/dev/shm/.ulsinst.my.cnf
  cleanfile="-C"
  echo '[client]
user=root' >$mysqlinstdefaults
  chmod 600 $mysqlinstdefaults
  if [[ -n "$dbpwd" ]]; then
    echo "password=$dbpwd" >>$mysqlinstdefaults
  fi
fi
#
if ! mysql --defaults-extra-file=${mysqldefaults:-$mysqlinstdefaults} -e "show databases" >/dev/null; then
  echo "Keine Verbindung zur Datenbank oder Kennwort falsch."
  exit 1
fi
#
if mysql --defaults-extra-file=${mysqldefaults:-$mysqlinstdefaults} -e "show variables like 'sql_mode'" | grep -q 'STRICT_TRANS_TABLES'; then
  echo "Error: 'STRICT_TRANS_TABLES' im sql_mode gefunden."
  exit 1
fi
#
if echo "show databases" | mysql --defaults-extra-file=${mysqldefaults:-$mysqlinstdefaults} | grep -q '^logging$'; then
  echo "Die Datenbank "logging" existiert schon!"
  if [[ -z "$force" ]]; then
    echo "Installation wird abgebrochen"
    exit 1
  elif [[ -z "$allwaysyes" ]]; then
    echo
    echo "Datenbank wird überschrieben"
    echo -n "Return druecken "
    read dummy
    echo
  fi
fi
echo
echo "Hilfsprogramme kompilieren"
#
mkdir /var/tmp/ulsinst
for s in src/*; do
  (cp -a $s /var/tmp/ulsinst/ && cd /var/tmp/ulsinst/${s#src/} && make)
  RET=$?
  [[ $RET != 0 ]] && echo "Fehler beim Kompilieren, Installation wird abgebrochen!" && exit 1
done
#
echo
if getent group ulsuser >/dev/null; then
  echo "Gruppe 'ulsuser' schon vorhanden"
  ulsgid=$(getent group ulsuser | cut -d : -f 3)
else
  echo "Es wird die Gruppe 'usluser' fuer die user 'uls' und 'wwwuls' angelegt"
  if [[ -z "$ulsgid$defaultids" ]]; then
    echo -n "gid fuer die Gruppe 'ulsuser' [500]? "
    read ulsgid
    echo
  fi
  if [[ -z "$ulsgid" ]]; then
    ulsgid=500
  fi
  groupadd -g $ulsgid ulsuser
fi
echo
if getent passwd uls >/dev/null; then
  echo "User 'uls' schon vorhanden"
else
  echo "Es wird der User angelegt, unter dem das ULS-System laeuft"
  if [[ -z "$ulsuid$defaultids" ]]; then
    echo -n "uid fuer den User 'uls' [500]? "
    read ulsuid
    echo
  fi
  if [[ -z "$ulsuid" ]]; then
    ulsuid=500
  fi
  useradd -c "Universal Logging System" -s /bin/bash -g $ulsgid -u $ulsuid uls
fi
echo
if getent passwd wwwuls >/dev/null; then
  echo "User 'wwwuls' schon vorhanden"
else
  if [[ -z "$wwwulsuid$defaultids" ]]; then
    echo "Es wird der User angelegt, unter dem der WWW-Server des ULS-Systems laeuft"
    echo -n "uid fuer den User 'wwwuls' [501]? "
    read wwwulsuid
    echo
  fi
  if [[ -z "$wwwulsuid" ]]; then
    wwwulsuid=501
  fi
  useradd -c "Universal Logging System WWW" -s /bin/bash -g $ulsgid -u $wwwulsuid wwwuls
fi
#
if [[ ! -d ~uls ]]; then
  mkdir -p ~uls
  chown uls:ulsuser ~uls
fi
#
echo
echo "Software auspacken"
#
mkdir -p "$ULSBASE"
#
tar xfJ uls_soft*.txz -C "$ULSBASE"
chmod 775 "$ULSBASE/uls/log"
#
echo
echo "Hilfsprogramme installieren"
#
for s in src/*; do
  (cd $s && make install)
  RET=$?
  [[ $RET != 0 ]] && echo "Fehler beim Installieren, Installation wird abgebrochen!" && exit 1
done
#
./installulsdb.sh "${mysqldefaults:-$mysqlinstdefaults}"
#
echo "Postinstallscript"
./postinst.sh "$ULSBASE" "${mysqldefaults:-$mysqlinstdefaults}"
#
echo
if [[ -f /etc/ulsserver.conf ]]; then
  echo "uls Konfig schon vorhanden."
else
  unset MYSQLSOCKET
  if [[ -n "$mysqldefaults" ]]; then
    MYSQLSOCKET=$(awk -F ' *= *' '$1 == "socket" {print $2}' "$mysqldefaults")
  fi
  echo "uls Konfig einrichten."
  if [[ -n "$MYSQLSOCKET$MYSQL_UNIX_PORT" ]]; then
    sed "s'%ULSBASE%'$ULSBASE'g;s'%MYFQDN%'$(hostname -f)'g;/^# export MYSQL_UNIX_PORT=/aexport MYSQL_UNIX_PORT=${MYSQLSOCKET:-$MYSQL_UNIX_PORT}" etc/ulsserver.conf >/etc/ulsserver.conf
  else
    sed "s'%ULSBASE%'$ULSBASE'g;s'%MYFQDN%'$(hostname -f)'g" etc/ulsserver.conf >/etc/ulsserver.conf
  fi
  chmod 644 /etc/ulsserver.conf
fi
#
echo "SSL Konfiguration"
./uls_certs.sh "$ULSBASE"
echo
if [[ -f /etc/cron.d/ulsserver ]]; then
  echo "Crontab fuer uls schon vorhanden."
else
  echo "Crontab fuer uls einrichten."
  sed "s'%ULSBASE%'$ULSBASE'g" etc/cron.d/ulsserver >/etc/cron.d/ulsserver
fi
#
echo
if [[ -f /etc/unix2web/uls.conf ]]; then
  echo "unix2web ist schon konfiguriert"
else
  mkdir -p /etc/unix2web
  cp etc/unix2web/uls.conf /etc/unix2web/uls.conf
fi
if [[ -f /etc/unix2web/uls80.conf ]]; then
  echo "unix2web 80 ist schon konfiguriert"
else
  cp etc/unix2web/uls80.conf /etc/unix2web/uls80.conf
fi
NODENAME=$(hostname -f)
if [[ -f /etc/unix2web/uls-https.conf ]]; then
  echo "unix2web HTTPS ist schon konfiguriert"
else
  sed "/^SSL_/s/ULS-server-/ULS-${NODENAME%%.*}-/" etc/unix2web/uls-https.conf >/etc/unix2web/uls-https.conf
fi
if [[ -f /etc/unix2web/uls443.conf ]]; then
  echo "unix2web 443 ist schon konfiguriert"
else
  sed "/^SSL_/s/ULS-server-/ULS-${NODENAME%%.*}-/" etc/unix2web/uls443.conf >/etc/unix2web/uls443.conf
fi
#
if which u2w_passwd &>/dev/null; then
  NEWPWD=$(tr -Cd '[:alnum:],._+=-' </dev/urandom | head -c 12)
  sed -i "s@^ULSREPPWD='getreport'@ULSREPPWD='$NEWPWD'@" "$ULSDIR/ulsreport.conf"
  u2w_passwd -p "$NEWPWD" "$ULSDIR/passwd.dat" doreport
fi
#
echo "unix2web starten"
/etc/init.d/unix2web enableall
systemctl start unix2web@* --all
#
echo
echo "Die Installation ist beendet."
echo
#
echo "Nun kann ULS ueber einen Browser aufgerufen werden: https://<DieserServer>/"
echo "Als erstes sollte man sich als User 'admin' mit Kennwort 'NewAdmin' anmelden,"
echo "um Verfahren, User und Server einzurichten."
echo
#
if [[ "$CLEANFILE" = "-C" ]]
 then
  rm "$MYSQLINSTDEFAULTS"
fi
