#!/bin/bash
#
# uls_certs.sh
#
# Create ULS CA and Server Cert
##################################
#
ULSDIR="$1/uls"
#
if [[ ! -d "$ULSDIR" ]]; then
  echo "Error: Missing dir '$ULSDIR'"
  exit 1
fi
#
WD=$(dirname $0)
#
NODENAME=$(hostname -f)
#
if [[ -f $WD/certs.conf ]]; then
  . $WD/certs.conf
fi
#
mkdir -p $ULSDIR/certs/private $ULSDIR/certs/certs
#
if [[ -f $ULSDIR/certs/private/ULS-cakey.pem || -f $ULSDIR/certs/private/ULS-${NODENAME%%.*}-key.pem || -f $ULSDIR/certs/certs/ULS-${NODENAME%%.*}-cert.pem ]]; then
  echo "ULS Zertifikation schon vorhanden"
else
  PWDDAT="$ULSDIR/certs/capwd.dat"
  tr -dc '0-9!@%a-zA-Z_,.;:+-' </dev/urandom | head -c20 >$PWDDAT
  chmod og-rwx $PWDDAT
  CAPWD="$(<$PWDDAT)"
  #
  openssl genrsa -aes256 -passout "pass:$CAPWD" -out $ULSDIR/certs/private/ULS-cakey.pem 2048
  openssl req -new -x509 -days 3650 -batch -subj "/C=$DE/ST=$STATE/L=$CITY/O=$COMPANY/OU=$OU/CN=ULS-CA-$(hostname)/emailAddress=$EMAIL" -passin "pass:$CAPWD" -key $ULSDIR/certs/private/ULS-cakey.pem -out $ULSDIR/certs/ULS-ca.pem -set_serial 1
  chmod 444 $ULSDIR/certs/ULS-ca.pem
  #
  touch $ULSDIR/certs/index.txt
  echo 01 >$ULSDIR/certs/serial
  #
  openssl req -new -newkey rsa:2048 -batch -subj "/C=$DE/ST=$STATE/L=$CITY/O=$COMPANY/OU=$OU/CN=$NODENAME/emailAddress=$EMAIL" -out $ULSDIR/certs/certs/ULS-${NODENAME%%.*}-csr.pem -nodes -keyout $ULSDIR/certs/private/ULS-${NODENAME%%.*}-key.pem
  chmod 400 $ULSDIR/certs/private/ULS-${NODENAME%%.*}-key.pem
  #
  openssl x509 -req -passin "pass:$CAPWD" -in $ULSDIR/certs/certs/ULS-${NODENAME%%.*}-csr.pem -out $ULSDIR/certs/certs/ULS-${NODENAME%%.*}-cert.pem -CA $ULSDIR/certs/ULS-ca.pem -CAkey $ULSDIR/certs/private/ULS-cakey.pem -CAserial $ULSDIR/certs/serial -days 3650
  #
  chmod 444 $ULSDIR/certs/certs/ULS-${NODENAME%%.*}-cert.pem
  rm $ULSDIR/certs/certs/ULS-${NODENAME%%.*}-csr.pem
fi
#
[[ -f $ULSDIR/certs/dhparam2048.pem ]] || openssl dhparam -out $ULSDIR/certs/dhparam2048.pem 2048
