#!/bin/bash
#
# call_mail_ret.sh
#
# send mail with attachments (mailreports)
#
# usage: call_mail_ret.sh <address> <cc> <returnaddr> <subject> <bodyfile> <attachments>
#
# Variablen in /etc/ulsserver.conf
#   MAILREP_BODYHEADER - insert text before mailbody ({SENDER} in text stands for user/group mailaddress
#   MAILREP_SENDER     - mailaddress of sender - not E-mail of user or group
#   MAILREP_NOSENDER   - use address of system as sender
###################################################################################
#
# Wenn es keine Anhaenge gibt und der Body leer ist, dann Ende
if [[ -z "$6" && ! -s "$5" ]]; then
  exit 0
fi
#
unset MAILREP_SENDER MAILREP_BODYHEADER MAILREP_NOSENDER
if [[ -f /etc/ulsserver.conf ]]; then
  . /etc/ulsserver.conf
fi
#
if [[ -n "$MAILREP_NOSENDER" ]]; then
  unset MAILREP_SENDER
elif [[ -z "$MAILREP_SENDER" ]]; then
  MAILREP_SENDER="$3"
fi
#
if [[ -n "$MAILREP_BODYHEADER" ]]; then
  mailbody="$(sed "s|{SENDER}|${3//|/}|g" <<<"$MAILREP_BODYHEADER")

$(<$5)"
else
  mailbody="$(<$5)"
fi
SUBJECT=$(echo $4)
MAIL=`which mail`
if [[ -z "$MAIL" ]]; then
  MAIL=`which mailx`
  if [[ -z "$MAIL" ]]; then
    MAIL=`which s-nail`
  fi
fi
if [[ -n "$MAIL" ]]; then
  MAILPRG=`readlink -e "$MAIL"`
  if [[ "$MAILPRG" == *s-nail || "$MAILPRG" == *mailx* && "$MAILPRG" != *bsd-mailx ]]; then
    if [[ -n "$2" ]]; then
      if [[ -n "$MAILREP_SENDER" ]]; then
        eval '$MAILPRG -c "$2" -r "$MAILREP_SENDER" -s "$SUBJECT" '"$6"' "$1" <<<"$mailbody"'
      else
        eval '$MAILPRG -c "$2" -s "$SUBJECT" '"$6"' "$1" <<<"$mailbody"'
      fi
    elif [[ -n "$MAILREP_SENDER" ]]; then
      eval '$MAILPRG -r "$MAILREP_SENDER" -s "$SUBJECT" '"$6"' "$1" <<<"$mailbody"'
    else
      eval '$MAILPRG -s "$SUBJECT" '"$6"' "$1" <<<"$mailbody"'
    fi
  else
    if [[ -n "$2" ]]; then
      eval '$MAILPRG -c "$2" -s "$SUBJECT" "$1" <<<"$mailbody"'
    else
      eval '$MAILPRG -s "$SUBJECT" "$1" <<<"$mailbody"'
    fi
  fi
else
  echo "ERROR: No mail program found."
fi
