/****************************************/
/* file2ulsdb.c                         */
/*                                      */
/* Programm zur Einsortierung der Werte */
/* in die ULS-Datenbank                 */
/****************************************/

#include "file2ulsdb.h"


MYSQL mh;

char zeile[MAX_ZEILENLAENGE], *z;
char folgezeile[MAX_ZEILENLAENGE];

char lastdatetime[MAX_DATETIME_LEN], lasttz[MAX_TZ_LEN],
     lastserver[MAX_SERVER_LEN], lastsection[MAX_SECTION_LEN],
     lastteststep[MAX_TESTSTEP_LEN];

long long serverid;

char *def_mysql_charset = ULS_MYSQL_CHARACTERSET;
int charset_changed = 0;


/***************************************************************************************/
/* char *strcpyn(char *out, char *in, unsigned long n)                                 */
/*              char *out: Ausgabezeile                                                */
/*              char *in : Eingabezeile                                                */
/*              long n   : Maximal n Zeichen                                           */
/*      strcpyn kopiert in nach out                                                    */
/***************************************************************************************/
char *strcpyn(char *out, char *in, unsigned long n)
{ long i;

  i = 1;
  while( *in && i++ < n )
    *out++ = *in++;
  *out = '\0';
  return out;
}


/***************************************************************************************/
/* char *strcpynm(char *out, char *in, unsigned long n, unsigned long m)               */
/*              char *out: Ausgabezeile                                                */
/*              char *in : Eingabezeile                                                */
/*              long n   : Maximal n Zeichen                                           */
/*              long m   : und Maximal m Zeichen                                       */
/*      strcpyn kopiert in nach out                                                    */
/***************************************************************************************/
char *strcpynm(char *out, char *in, unsigned long n, unsigned long m)
{ if( n < m+1 )
    return strcpyn(out, in, n);
  else
    return strcpyn(out, in, m+1);
}


/***************************************************************************************/
/* char *get_ct(char *file)                                                            */
/*              char *file: Dateiname zur Bestimmung des Content-Types                 */
/*              return    : Liefert Zeiger auf den Content-Type der Datei file         */
/***************************************************************************************/
char *get_ct(char *file)
{ int i;
  char *p;

  i = 0;
  if( NULL != (p = strrchr(file, '.')) )
  { if( *p )
      p++;
    while( *content_types[i][0] && strcmp(content_types[i][0], p) )
      i++;
    if( *content_types[i][0] )
      return content_types[i][1];
    else
    { return "application/x-";
    }
  }
  else
    return "text/plain";
}


void mysql_quote(char *out, char *in)
{ char *p, *q;

  p = in;
  q = out;
  while( *p )
  { if( *p == '\\' || *p == '\'' )
      *q++ = '\\';
    *q++ = *p++;
  }
  *q = '\0';
}

char *gen_query(char *query, unsigned long n, char *q, ...)
{ char *s, *o, *p;
  va_list ap;
  long long l;
  my_ulonglong u;

  va_start(ap, q);

  s = q;
  o = query;
  while( *s && o-query < n )
  { if( *s == '%' )
    { s++;
      switch(*s)
      { case 's': p = va_arg(ap, char *);
                  while( *p && o-query < n )
                    *o++ = *p++;
                  break;
        case 'S': p = va_arg(ap, char *);
                  while( *p && o-query < n-1 )
                  { if( *p == '\\' || *p == '\'' )
                      *o++ = '\\';
                    *o++ = *p++;
                  }
                  break;
        case 'D': p = va_arg(ap, char *);
                  if( strchr(p, '-') )
                  { while( *p && o-query < n )
                      *o++ = *p++;
                  }
                  else
                  { int j, m, t;

                    sscanf(p, "%d.%d.%d", &t, &m, &j);
                    if( j < 100 )
                      j += 2000;
                    if( m < 1 || 12 < m )
                      m = 1;
                    if( t < 1 || 31 < t )
                      t = 1;
                    o += snprintf(o, MAX_QUERY_LEN-(o-query), "%04d-%02d-%02d", j, m, t);
                  }
                  break;
        case 'L': l = va_arg(ap, long long);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%lld", l);
                  break;
        case 'U': u = va_arg(ap, my_ulonglong);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%llu", u);
                  break;
        default:  *o++ = *s;
      }
      s++;
    }
    else
      *o++ = *s++;
  }
  *o = '\0';

  va_end(ap);

  return o;
}


int do_mysql_query(MYSQL *mh, char *q, ...)
{ char query[MAX_QUERY_LEN];
  char *s, *o, *p;
  long long l;
  int i;
  my_ulonglong u;
  va_list ap;

  va_start(ap, q);

  s = q;
  o = query;
  while( *s && o-query < MAX_QUERY_LEN )
  { if( *s == '%' )
    { s++;
      switch(*s)
      { case 's': p = va_arg(ap, char *);
                  while( *p && o-query < MAX_QUERY_LEN )
                    *o++ = *p++;
                  break;
        case 'S': p = va_arg(ap, char *);
                  while( *p && o-query < MAX_QUERY_LEN-1 )
                  { if( *p == '\\' || *p == '\'' )
                      *o++ = '\\';
                    *o++ = *p++;
                  }
                  break;
        case 'D': p = va_arg(ap, char *);
                  if( strchr(p, '-') )
                  { while( *p && o-query < MAX_QUERY_LEN )
                      *o++ = *p++;
                  }
                  else
                  { int j, m, t;

                    sscanf(p, "%d.%d.%d", &t, &m, &j);
                    if( j < 100 )
                      j += 2000;
                    if( m < 1 || 12 < m )
                      m = 1;
                    if( t < 1 || 31 < t )
                      t = 1;
                    o += snprintf(o, MAX_QUERY_LEN-(o-query), "%04d-%02d-%02d", j, m, t);
                  }
                  break;
        case 'L': l = va_arg(ap, long long);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%lld", l);
                  break;
        case 'd': i = va_arg(ap, int);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%d", i);
                  break;
        case 'U': u = va_arg(ap, my_ulonglong);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%llu", u);
                  break;
        default:  *o++ = *s;
      }
      s++;
    }
    else
      *o++ = *s++;
  }

  va_end(ap);

  if( o-query < MAX_QUERY_LEN )
  { *o = '\0';
#ifdef DEBUG
  fprintf(stderr, "mysql_query: %s.\n", query);
#endif
    if( mysql_query(mh, query) )
    { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(mh));
      fprintf(stderr, "query: %s\n", query);
      return 1;
    }
    else
      return 0;
  }
  else
    return 1;
}

my_ulonglong mysql_query_int(MYSQL *mh, char *q, ...)
{ char query[MAX_QUERY_LEN];
  char *s, *o, *p;
  long long l;
  my_ulonglong u;
  va_list ap;
  MYSQL_RES *mres;
  MYSQL_ROW mrow;

  va_start(ap, q);

  l = 0;
  s = q;
  o = query;
  while( *s && o-query < MAX_QUERY_LEN )
  { if( *s == '%' )
    { s++;
      switch(*s)
      { case 's': p = va_arg(ap, char *);
                  while( *p && o-query < MAX_QUERY_LEN )
                    *o++ = *p++;
                  break;
        case 'S': p = va_arg(ap, char *);
                  while( *p && o-query < MAX_QUERY_LEN-1 )
                  { if( *p == '\\' || *p == '\'' )
                      *o++ = '\\';
                    *o++ = *p++;
                  }
                  break;
        case 'D': p = va_arg(ap, char *);
                  if( strchr(p, '-') )
                  { while( *p && o-query < MAX_QUERY_LEN )
                      *o++ = *p++;
                  }
                  else
                  { int j, m, t;

                    sscanf(p, "%d.%d.%d", &t, &m, &j);
                    if( j < 100 )
                      j += 2000;
                    if( m < 1 || 12 < m )
                      m = 1;
                    if( t < 1 || 31 < t )
                      t = 1;
                    o += snprintf(o, MAX_QUERY_LEN-(o-query), "%04d-%02d-%02d", j, m, t);
                  }
                  break;
        case 'L': l = va_arg(ap, long long);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%lld", l);
                  break;
        case 'U': u = va_arg(ap, my_ulonglong);
                  o += snprintf(o, MAX_QUERY_LEN-(o-query), "%llu", u);
                  break;
        default:  *o++ = *s;
      }
      s++;
    }
    else
      *o++ = *s++;
  }

  va_end(ap);

  if( o-query < MAX_QUERY_LEN )
  { *o = '\0';
#ifdef DEBUG
  fprintf(stderr, "mysql_query: %s.\n", query);
#endif
    if( mysql_query(mh, query) )
    { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(mh));
      fprintf(stderr, "query: %s\n", query);
      return 0;
    }
    l = 0;

    if( (mres = mysql_store_result(mh)) )
    { if( mysql_num_fields(mres) > 0 )
      { if( (mrow = mysql_fetch_row(mres)) )
        { unsigned long *lengths;

          lengths = mysql_fetch_lengths(mres);
          strcpynm(query, mrow[0] ? mrow[0] : "", MAX_QUERY_LEN, lengths[0]);
          l = atoll(query);
        }
      }
      mysql_free_result(mres);
    }
  }
#ifdef DEBUG
  fprintf(stderr, "return: %lld.\n", l);
#endif
  return l;
}

char *ip = "";
char *ssl = "";
char mode[MAX_MODE_LEN];
char datetime[MAX_DATETIME_LEN];
short microsecondflag = 0;
char tz[MAX_TZ_LEN];
char server[MAX_SERVER_LEN];
char section[MAX_SECTION_LEN];
char teststep[MAX_TESTSTEP_LEN];
char detail[MAX_DETAIL_LEN];
char wert[MAX_WERT_LEN];
char einheit[MAX_EINHEIT_LEN];
char accessright[MAX_ACCESS_LEN];
char extra[MAX_EXTRA_LEN];

char p_datetime[MAX_DATETIME_LEN];
char p_tz[MAX_TZ_LEN];
char p_server[MAX_SERVER_LEN];
char p_section[MAX_SECTION_LEN];
char p_teststep[MAX_TESTSTEP_LEN];
char p_detail[MAX_DETAIL_LEN];
char p_stopdatetime[MAX_DATETIME_LEN];
char p_stoptz[MAX_TZ_LEN];
char p_einheit[MAX_EINHEIT_LEN];
char p_accessright[MAX_ACCESS_LEN];
char p_remark[MAX_EXTRA_LEN];

#ifdef DEBUGSIG
#include <signal.h>
/***************************************************************************************/
/* void sigsegv_handler(int signum)                                                    */
/*                      int signum: ausloesendes Signal                                */
/***************************************************************************************/
void sigsegv_handler(int signum)
{ FILE *ptr;

  if( NULL != (ptr = fopen("/tmp/file2ulsdb_sigseg_err.log", "a")) )
  { fprintf(ptr, "ERROR sigseg from %s (%s) in line: %s", server, ip, zeile);
    fclose(ptr);
    fprintf(stderr, "ERROR %d from %s (%s) in line: %s", 511, server, ip, zeile);
    exit(511);
  }
}
#endif

int store_blob(char *path, char *name, char *charset)
{ char wtab[LENWTAB];
  char *p, *puffer, zeile[MAX_ZEILENLAENGE];
  MYSQL_RES *mres;
  MYSQL_ROW mrow;
  int hd_in, ret;
  long nb, i;
  my_ulonglong insid;
#ifdef _LARGEFILE64_SOURCE
  struct stat64 stat_buf;                            /* fuer stat-Aufruf               */
#else
  struct stat stat_buf;                              /* fuer stat-Aufruf               */
#define stat64(a,b) stat(a,b)
#define open64(a,b) open(a,b)
#endif

  if( strcmp(server, lastserver) )
  { if( *ssl )
      serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                 server, ssl);
    else
      serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                 server, ip);
    strcpy(lastserver, server);
    if( ! serverid )
      return ERROR_NO_SERVER;
  }
  else if( ! serverid )
    return ERROR_NO_SERVER_AGAIN;

  if( !*datetime && !*lastdatetime )
    return ERROR_NO_DATETIME;
  else if( strcmp(datetime, lastdatetime) )
  { if( mysql_query_int(&mh, "select (@dt := '%S') > now() + interval 48 hour", datetime) )
    { *lastdatetime = '\0';
      return ERROR_DATETIME_IN_FUTURE;
    }
    strcpy(lastdatetime, datetime);
  }

  ret = 0;

  if( (p = strrchr(path, '/')) )
    path = p+1;

  if( (!stat64(path, &stat_buf)                      /* Ist Datei vorhanden?           */
       && (S_IFREG & stat_buf.st_mode) == S_IFREG))  /* und normale Datei              */
  { if( 0 <= (hd_in = open64(path, O_RDONLY)) )      /* Datei oeffnen                  */
    { char *q;

      if( NULL == (q = strrchr(name, '/')) )
        q = name;

      if( strcmp(section, lastsection) )
      { if( !*section || do_mysql_query(&mh, "call newsec('%S', @s)", section) )
          return ERROR_NO_STDE;
        strcpy(lastsection, section);
      }
      else if( !*lastsection )
        return ERROR_NO_STDE;

      if( strcmp(teststep, lastteststep) )
      { if( !*teststep || do_mysql_query(&mh, "call newtst('%S', @t)", teststep) )
          return ERROR_NO_STDE;
        strcpy(lastteststep, teststep);
      }
      else if( !*lastteststep )
        return ERROR_NO_STDE;

      if( *mode == 'F' )
      { if( do_mysql_query(&mh, "call getunit('BLOB', @ei)") )
          return ERROR_NO_STDE;
      }
      else if( *mode == 'I' )
      { if( do_mysql_query(&mh, "call getunit('IMAGE', @ei)") )
          return ERROR_NO_STDE;
      }
      else
        return 1;

      if( !*detail || do_mysql_query(&mh, "call newdet('%S', @ei, @d)", detail) )
        return ERROR_NO_STDE;

      if( do_mysql_query(&mh, "select getacc('%S') into @aid", accessright) )
        return ERROR_NO_ACCESS;

      if( mysql_query(&mh, "call newwid(@sr, @s, @t, @d, @aid, @wi)") )
        return ERROR_NO_WID;

      if( mysql_query(&mh, "select wtab from einheiten where ID=@ei") )
        return ERROR_NO_WTAB;
      else
      { if( (mres = mysql_store_result(&mh)) )
        { if( mysql_num_fields(mres) > 0 )
          { if( (mrow = mysql_fetch_row(mres)) )
            { unsigned long *lengths;

              lengths = mysql_fetch_lengths(mres);
              strcpynm(wtab, mrow[0] ? mrow[0] : "", LENWTAB, lengths[0]);
            }
          }
          mysql_free_result(mres);

          if( stat_buf.st_size > 8*1024*1024 )
          { int concatflag;
            char *p;

            if( NULL == (puffer = malloc(MYSQL_PUFFERSIZE)) )
            { fprintf(stderr, "Error on allocating memory.\n");
              close(hd_in);
              return true;
            }

            concatflag = false;

            p = puffer + snprintf(puffer, MYSQL_PUFFERSIZE, "set @b=BINARY'");
            while( 0 < (nb = read(hd_in, zeile, MAX_ZEILENLAENGE)) )
            { for( i = 0; i < nb; i++ )
              { switch( zeile[i] )
                { case '\0': *p++ = '\\';
                             *p++ = '0';
                             break;
                  case '\\':
                  case '\'': *p++ = '\\';
                  default:   *p++ = zeile[i];
                }
                if( p-puffer >= MYSQL_PUFFERSIZE-100 )
                { if( concatflag )
                    strcpyn(p, "')", MYSQL_PUFFERSIZE - (p-puffer));
                  else
                  { strcpyn(p, "'", MYSQL_PUFFERSIZE - (p-puffer));
                    concatflag = true;
                  }
                  if( mysql_query(&mh, puffer) )
                  { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
                    close(hd_in);
                    free(puffer);
                    return true;
                  }
                  p = puffer + snprintf(puffer, MYSQL_PUFFERSIZE, "set @b=concat(@b,BINARY'");
                  p = puffer + strlen(puffer);
                }
              }
            }
            if( concatflag )
              strcpyn(p, "')", MYSQL_PUFFERSIZE - (p-puffer));
            else
              strcpyn(p, "'", MYSQL_PUFFERSIZE - (p-puffer));
            if( mysql_query(&mh, puffer) )
            { fprintf(stderr, "mysql_query error: %s.\n",
                        mysql_error(&mh));
              free(puffer);
              close(hd_in);
              return true;
            }
            free(puffer);
          }
          else
          { if( NULL != (puffer = malloc(100+2*stat_buf.st_size)) )
            { p = puffer + snprintf(puffer, MYSQL_PUFFERSIZE, "set @b=BINARY'");
              while( 0 < (nb = read(hd_in, zeile, MAX_ZEILENLAENGE)) )
              { for( i = 0; i < nb; i++ )
                { switch( zeile[i] )
                  { case '\0': *p++ = '\\';
                               *p++ = '0';
                               break;
                    case '\\':
                    case '\'': *p++ = '\\';
                    default:   *p++ = zeile[i];
                  }
                }
              }
              *p++ = '\'';
              *p = '\0';
              if( mysql_query(&mh, puffer) )
              { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
                ret = ERROR_STORE_BLOB;
              }
              free(puffer);
            }
            else
            { fprintf(stderr, "Error on allocation buffer.\n");
              close(hd_in);
              return true;
            }
          }

          if( mode[0] == 'F' && mode[1] == 'A' )
          { insid = mysql_query_int(&mh, "select ID from %s "
                "where werteindizesID=@wi and dateiname='%S' and mime='%s' "
                "and datum=@dt order by ID limit 1",
                wtab, q, get_ct(q));
          }
          else
            insid = 0;

          if( insid > 0 )
            do_mysql_query(&mh, "call appendblob('%U')", insid);
          else
          { do_mysql_query(&mh, "call storeblob(@bid)");
            do_mysql_query(&mh, "insert into %s set werteindizesID = @wi, datum = @dt, datenID = @bid, dateiname = '%S', mime = '%s', charset = '%S'",
                     wtab, q, get_ct(q), charset);
          }
        }
      }
      close(hd_in);

      do_mysql_query(&mh, "update letztewerte set lastwert='%S', llastdatum=lastdatum, lastdatum=@dt where ID=@wi and @dt >= lastdatum",
               q);

      mysql_query(&mh, "call wzeiten(@wi, @sr, @s, @t, @d, @aid, @dt)");
    }
    else
    { fprintf(stderr, "Error on opening file %s.\n", path);
      return ERROR_FILE_OPEN;
    }
  }
  else
  { fprintf(stderr, "Error on locating file %s.\n", path);
    return ERROR_FILE_NOT_FOUND;
  }
  return ret;
}

int store_value(void)
{ char wtab[LENWTAB];
  MYSQL_RES *mres;
  MYSQL_ROW mrow;

  if( strcmp(server, lastserver) )
  { if( *ssl )
      serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                 server, ssl);
    else
      serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                 server, ip);
    strcpy(lastserver, server);
    if( ! serverid )
      return ERROR_NO_SERVER;
  }
  else if( ! serverid )
    return ERROR_NO_SERVER_AGAIN;

  if( !*datetime && !*lastdatetime )
    return ERROR_NO_DATETIME;
  else if( strcmp(datetime, lastdatetime) )
  { if( mysql_query_int(&mh, "select (@dt := '%S') > now() + interval 48 hour", datetime) )
    { *lastdatetime = '\0';
      return ERROR_DATETIME_IN_FUTURE;
    }
    strcpy(lastdatetime, datetime);
  }

  if( strcmp(section, lastsection) )
  { if( !*section || do_mysql_query(&mh, "call newsec('%S', @s)", section) )
      return ERROR_NO_STDE;
    strcpy(lastsection, section);
  }
  else if( !*lastsection )
    return ERROR_NO_STDE;

  if( strcmp(teststep, lastteststep) )
  { if( !*teststep || do_mysql_query(&mh, "call newtst('%S', @t)", teststep) )
      return ERROR_NO_STDE;
    strcpy(lastteststep, teststep);
  }
  else if( !*lastteststep )
    return ERROR_NO_STDE;

  if( *mode == 'C' )
  { if( do_mysql_query(&mh, "call getunit('COUNTER', @ei)") )
      return ERROR_NO_STDE;
    strcpy(wtab, "cwerte");
  }
  else
  { if( microsecondflag )
    { if( do_mysql_query(&mh, "call getmunitt('%S', @ei, @wtab)", einheit) )
        return ERROR_NO_STDE;
    }
    else
    { if( do_mysql_query(&mh, "call getunitt('%S', @ei, @wtab)", einheit) )
        return ERROR_NO_STDE;
    }
    if( mysql_query(&mh, "select @wtab") )
    { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
      return ERROR_NO_WTAB;
    }
    else
    { if( (mres = mysql_store_result(&mh)) )
      { if( mysql_num_fields(mres) > 0 )
        { if( (mrow = mysql_fetch_row(mres)) )
          { unsigned long *lengths;

            lengths = mysql_fetch_lengths(mres);
            strcpynm(wtab, mrow[0] ? mrow[0] : "", LENWTAB, lengths[0]);
          }
        }
        mysql_free_result(mres);
      }
    }
  }

  if( !*detail || do_mysql_query(&mh, "call newdet('%S', @ei, @d)", detail) )
    return ERROR_NO_STDE;

  if( do_mysql_query(&mh, "select getacc('%S') into @aid", accessright) )
    return ERROR_NO_ACCESS;

  if( mysql_query(&mh, "call newwid(@sr, @s, @t, @d, @aid, @wi)") )
    return ERROR_NO_WID;

  if( *mode == 'N' )
  { if( mysql_query_int(&mh,
          "select ID from letztewerte where ID=@wi and lastwert='%S'",
             wert) > 0 )
      return 0;
  }
  else if( *mode == 'U' )
  { if( mysql_query_int(&mh,
          "select count(*) from %s where werteindizesID=@wi and wert='%S' and datum=@dt",
             wtab, wert) > 0 )
      return 0;
  }
  else if( *mode == 'O' )
  { int r;
    r = mysql_query_int(&mh,
          "select wert='%S' as r from %s where werteindizesID=@wi and datum=@dt order by wert!='%S' limit 1",
             wert, wtab, wert);
    if( r == 1 )
      return 0;
    else if( mysql_affected_rows(&mh) )
    { if( mysql_query_int(&mh, "select overwrite from access where ID = @aid") )
      { if( wtab[0] == 's' || wtab[0] == 't' )
        { do_mysql_query(&mh, "update %s set wert='%S' where werteindizesID=@wi and datum=@dt and id=0",
                         wtab, wert);
          do_mysql_query(&mh, "update letztewerte set lastwert='%S' where ID=@wi and lastdatum=@dt and sid=0",
                 wert);
        }
        else
        { do_mysql_query(&mh, "update %s set wert='%S' where werteindizesID=@wi and datum=@dt",
                         wtab, wert);
          do_mysql_query(&mh, "update letztewerte set lastwert='%S' where ID=@wi and lastdatum=@dt",
                 wert);
        }
      }
      return 0;
    }
  }
  else if( *mode == 'E' )
  { if( mysql_query_int(&mh,
          "select ID from letztewerte where ID=@wi and lastwert='%S' and lastdatum > @dt - interval '%S' hour_minute",
             wert, mode+1) > 0 )
      return 0;
  }
  else if( *mode == 'C' )
  { long long lw;

    if( !strcasecmp(wert, "reset") )
      mysql_query(&mh, "update letztewerte set lastwert=0, llastdatum=lastdatum, lastdatum=@dt where ID=@wi");
    else if( !strcasecmp(wert, "set") )
    { lw = atol(einheit);
      do_mysql_query(&mh, "update letztewerte set lastwert='%L', llastdatum=lastdatum, lastdatum=@dt where ID=@wi",
                     lw);
    }
    else if( !strcasecmp(wert, "add") )
    { lw = atol(einheit);
      do_mysql_query(&mh, "update letztewerte set lastwert=lastwert+'%L', llastdatum=lastdatum, lastdatum=@dt where ID=@wi",
                     lw);
    }
    else if( !strcasecmp(wert, "sub") )
    { lw = atol(einheit);
      do_mysql_query(&mh, "update letztewerte set lastwert=lastwert-'%L', llastdatum=lastdatum, lastdatum=@dt where ID=@wi",
                     lw);
    }
    else if( !strcasecmp(wert, "dec") )
      mysql_query(&mh, "update letztewerte set lastwert=lastwert-1, llastdatum=lastdatum, lastdatum=@dt where ID=@wi");
    else
      mysql_query(&mh, "update letztewerte set lastwert=lastwert+1, llastdatum=lastdatum, lastdatum=@dt where ID=@wi");

    mysql_query(&mh, "select lastwert into @nw from letztewerte where ID=@wi");
    if( do_mysql_query(&mh,
                       "insert ignore into %s set werteindizesID=@wi, wert=@nw, datum=@dt",
                       wtab) )
      return ERROR_WERT;
 
    mysql_query(&mh, "call wzeiten(@wi, @sr, @s, @t, @d, @aid, @dt)");
    return 0;
  }

  if( wtab[0] == 's' || wtab[0] == 't' )
  { do_mysql_query(&mh, "update letztewerte set lastwert='%S', sid = @sid := if(lastdatum=@dt, sid + 1, 0), llastdatum=lastdatum, lastdatum=@dt where ID=@wi and @dt >= lastdatum",
               wert);


    if( mysql_affected_rows(&mh) == 0 )
      do_mysql_query(&mh,
                     "set @sid = (select count(*) from %s where werteindizesID=@wi and datum=@dt)",
                     wtab);
    if( do_mysql_query(&mh,
                       "insert ignore into %s set werteindizesID=@wi, wert='%S', datum=@dt, id=@sid",
                       wtab, wert) )
      return ERROR_WERT;

  }
  else
  { do_mysql_query(&mh, "update letztewerte set lastwert='%S', llastdatum=lastdatum, lastdatum=@dt where ID=@wi and @dt >= lastdatum",
               wert);


    if( do_mysql_query(&mh,
                       "insert ignore into %s set werteindizesID=@wi, wert='%S', datum=@dt",
                       wtab, wert) )
      return ERROR_WERT;
  }
 
  mysql_query(&mh, "call wzeiten(@wi, @sr, @s, @t, @d, @aid, @dt)");

  return 0;
}


int store_teststep_remark(void)
{ my_ulonglong secid, tstid;

  if( *wert )
  { if( strcmp(server, lastserver) )
    { if( *ssl )
      serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                 server, ssl);
    else
      serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                 server, ip);
    strcpy(lastserver, server);
      if( ! serverid )
        return ERROR_NO_SERVER;
    }
    else if( ! serverid )
      return ERROR_NO_SERVER_AGAIN;

    if( !(secid = mysql_query_int(&mh, "select ID from section where bname='%S'",
                                  section)) )
      return ERROR_NO_STDE;

    if( !(tstid = mysql_query_int(&mh, "select ID from teststep where bname='%S'",
                                  teststep)) )
      return ERROR_NO_STDE;

    if( mysql_query_int(&mh, "select count(*) from sstbeschreibung "
                          "where serverID=@sr and sectionID='%U' and teststepID='%U'",
                        secid, tstid) > 0 )
    { if( do_mysql_query(&mh, "update sstbeschreibung set beschreibung='%S' "
                           "where serverID=@sr and sectionID='%U' and teststepID='%U'",
                         wert, secid, tstid) )
        return ERROR_SSTDOKU;
    }
    else
    { if( do_mysql_query(&mh, "insert into sstbeschreibung "
                            "set serverID=@sr, sectionID='%U', teststepID='%U', "
                            "beschreibung='%S'",
                         secid, tstid, wert) )
        return ERROR_SSTDOKU;
    }
  }
  return 0;
}

int store_section_remark(void)
{ my_ulonglong secid;

  if( *wert )
  { if( strcmp(server, lastserver) )
    { if( *ssl )
        serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                   server, ssl);
      else
        serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                   server, ip);
      strcpy(lastserver, server);
      if( ! serverid )
        return ERROR_NO_SERVER;
    }
    else if( ! serverid )
      return ERROR_NO_SERVER_AGAIN;

    if( !(secid = mysql_query_int(&mh, "select ID from section where name='%S'",
                                section)) )
      return ERROR_NO_STDE;

    if( mysql_query_int(&mh, "select count(*) from ssremarks "
                             "where serverID=@sr and sectionID='%U'",
                        secid) > 0 )
    { if( do_mysql_query(&mh, "update ssremarks set remark='%S' "
                           "where serverID=@sr and sectionID='%U'",
                         wert, secid) )
        return ERROR_SSTDOKU;
    }
    else
    { if( do_mysql_query(&mh, "insert into ssremarks "
                            "set serverID=@sr, sectionID='%U', remark='%S'",
                         secid, wert) )
        return ERROR_SSTDOKU;
    }
  }

  return 0;
}



int store_detail_remark(void)
{ my_ulonglong wid;

  if( *wert )
  { if( strcmp(server, lastserver) )
    { if( *ssl )
        serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                   server, ssl);
      else
        serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                   server, ip);
      strcpy(lastserver, server);
      if( ! serverid )
        return ERROR_NO_SERVER;
    }
    else if( ! serverid )
      return ERROR_NO_SERVER_AGAIN;

    if( !(wid = mysql_query_int(&mh,
                  "select w.ID from werteindizes as w, section as c, teststep as t, "
                   "teststepdetails as d, einheiten as e "
                   "where serverID=@sr and c.bname='%S' and t.bname='%S' "
                     "and e.bname='%S' and d.bname='%S' and d.einheitID=e.ID "
                     "and w.sectionID=c.ID and w.teststepID=t.ID "
                     "and w.teststepdetailsID=d.ID and w.accessID=getacc('%S')",
                  section, teststep, *einheit ? einheit : " ", detail,
                  accessright)) )
      return ERROR_NO_STDE;

    if( mysql_query_int(&mh, "select count(*) from beschreibung where ID='%L'",
                        wid) > 0 )
    { if( do_mysql_query(&mh, "update beschreibung set beschreibung='%S' where ID='%L'",
                         wert, wid) )
        return ERROR_SSTDOKU;
    }
    else
    { if( do_mysql_query(&mh, "insert into beschreibung set ID='%L', beschreibung='%S'",
                         wid, wert) )
        return ERROR_SSTDOKU;
    }
  }

  return 0;
}


int store_server_doku(char *name, char *beschreibung, char *rettime,
                      char *path, char *fname, char *charset)
{ char *p, *puffer, zeile[MAX_ZEILENLAENGE];
  int hd_in, ret;
  long nb, i;
#ifdef _LARGEFILE64_SOURCE
  struct stat64 stat_buf;                            /* fuer stat-Aufruf               */
#else
  struct stat stat_buf;                              /* fuer stat-Aufruf               */
#define stat64(a,b) stat(a,b)
#define open64(a,b) open(a,b)
#endif

  if( strcmp(server, lastserver) )
  { if( *ssl )
      serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                 server, ssl);
    else
      serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                 server, ip);
    strcpy(lastserver, server);
    if( ! serverid )
      return ERROR_NO_SERVER;
  }
  else if( ! serverid )
    return ERROR_NO_SERVER_AGAIN;

  if( !*datetime && !*lastdatetime )
    return ERROR_NO_DATETIME;
  else if( strcmp(datetime, lastdatetime) )
  { if( mysql_query_int(&mh, "select (@dt := '%S') > now() + interval 48 hour", datetime) )
    { *lastdatetime = '\0';
      return ERROR_DATETIME_IN_FUTURE;
    }
    strcpy(lastdatetime, datetime);
  }

  ret = 0;

  if( (!stat64(path, &stat_buf)                      /* Ist Datei vorhanden?           */
       && (S_IFREG & stat_buf.st_mode) == S_IFREG))  /* und normale Datei              */
  { if( 0 <= (hd_in = open64(path, O_RDONLY)) )      /* Datei oeffnen                  */
    { char *q;

      if( NULL == (q = strrchr(fname, '/')) )
        q = fname;

      do_mysql_query(&mh,
        "delete from serverdoku where serverID=@sr and name='%S'", name);

      if( *rettime && atoi(rettime) > 0 )
        do_mysql_query(&mh, "set @rettime = '%d'", atoi(rettime));
      else
        DOFKT(do_, mysql_query)(&mh, "select wert + 0 into @rettime from systemvars "
                                     "where name = 'StoretimeServerdocu'");
      /* groe Datei ggf. in mehreren Schritten speichern */
      if( stat_buf.st_size > 8*1024*1024 )
      { int concatflag;
        char *p;

        if( NULL == (puffer = malloc(MYSQL_PUFFERSIZE)) )
        { fprintf(stderr, "Error on allocating memory.\n");
          close(hd_in);
          return ERROR_SERVERDOKU;
        }

        concatflag = false;
        p = gen_query(puffer, MYSQL_PUFFERSIZE, "insert into serverdoku set "
                        "serverID=@sr, name='%S', dateiname='%S', mime='%s', "
                        "charset='%S', datum=@dt, beschreibung='%S', "
                        "deldate = @dt + interval @rettime day, daten='",
               name, q, get_ct(q), charset, beschreibung);
        while( 0 < (nb = read(hd_in, zeile, MAX_ZEILENLAENGE)) )
        { for( i = 0; i < nb; i++ )
          { switch( zeile[i] )
            { case '\0': *p++ = '\\';
                         *p++ = '0';
                         break;
              case '\\':
              case '\'': *p++ = '\\';
              default:   *p++ = zeile[i];
            }
            if( p-puffer >= MYSQL_PUFFERSIZE - 100 )
            { if( concatflag )
                strcpyn(p, "') where ID=last_insert_id()",
                        MYSQL_PUFFERSIZE - (p-puffer));
              else
              { strcpyn(p, "'", MYSQL_PUFFERSIZE - (p-puffer));
                concatflag = true;
              }
              if( mysql_query(&mh, puffer) )
              { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
                close(hd_in);
                free(puffer);
                return ERROR_SERVERDOKU;
              }
              p = strcpyn(puffer, "update serverdoku set daten=concat(daten,'",
                          MYSQL_PUFFERSIZE);
            }
          }
        }
        if( concatflag )
          strcpyn(p, "') where ID=last_insert_id()", MYSQL_PUFFERSIZE - (p-puffer));
        else
          strcpyn(p, "'", MYSQL_PUFFERSIZE - (p-puffer));
        if( mysql_query(&mh, puffer) )
        { fprintf(stderr, "mysql_query error: %s.\n",
                    mysql_error(&mh));
          free(puffer);
          close(hd_in);
          return ERROR_SERVERDOKU;
        }
        free(puffer);
      }
      else
      { if( NULL != (puffer = malloc(300+MAX_SECTION_LEN+MAX_TESTSTEP_LEN+MAX_WERT_LEN
                                     +2*stat_buf.st_size)) )
        { p = gen_query(puffer, 300+MAX_SECTION_LEN+MAX_TESTSTEP_LEN+MAX_WERT_LEN,
                        "insert into serverdoku set serverID=@sr, name='%S', "
                        "dateiname='%S', mime='%s', charset='%S', datum=@dt, beschreibung='%S', "
                        "deldate = @dt + interval @rettime day, daten='",
               name, q, get_ct(q), charset, beschreibung);

          while( 0 < (nb = read(hd_in, zeile, MAX_ZEILENLAENGE)) )
          { for( i = 0; i < nb; i++ )
            { switch( zeile[i] )
              { case '\0': *p++ = '\\';
                           *p++ = '0';
                           break;
                case '\\':
                case '\'': *p++ = '\\';
                default:   *p++ = zeile[i];
              }
            }
          }
          strcpyn(p, "'", MYSQL_PUFFERSIZE - (p-puffer));
          if( mysql_query(&mh, puffer) )
          { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
            ret = ret | ERROR_SERVERDOKU;
          }
          free(puffer);
        }
        else
        { fprintf(stderr, "Error on allocating buffer.\n");
          ret = ret | ERROR_SERVERDOKU;
        }
      }
      close(hd_in);
    }
    else
    { fprintf(stderr, "Error on opening file %s.\n", path);
      return ERROR_FILE_OPEN;
    }
  }
  else
  { fprintf(stderr, "Error on locating file %s.\n", path);
    return ERROR_FILE_NOT_FOUND;
  }

  return ret;
}


int store_server_doku_value(char *name, char *beschreibung, char *rettime,
                            char *wert, char *fname, char *charset)
{ int ret;
  char *q;

  if( strcmp(server, lastserver) )
  { if( *ssl )
      serverid = mysql_query_int(&mh, "select @sr:=getserveridx('%S', '%S')",
                                 server, ssl);
    else
      serverid = mysql_query_int(&mh, "select @sr:=getserveridi('%S', '%S')",
                                 server, ip);
    strcpy(lastserver, server);
    if( ! serverid )
      return ERROR_NO_SERVER;
  }
  else if( ! serverid )
    return ERROR_NO_SERVER_AGAIN;

  if( !*datetime && !*lastdatetime )
    return ERROR_NO_DATETIME;
  else if( strcmp(datetime, lastdatetime) )
  { if( mysql_query_int(&mh, "select (@dt := '%S') > now() + interval 48 hour", datetime) )
    { *lastdatetime = '\0';
      return ERROR_DATETIME_IN_FUTURE;
    }
    strcpy(lastdatetime, datetime);
  }

  ret = 0;

  if( NULL == (q = strrchr(fname, '/')) )
    q = fname;

  do_mysql_query(&mh,
    "delete from serverdoku where serverID=@sr and name='%S'", name);

  if( *rettime && atoi(rettime) > 0 )
    do_mysql_query(&mh, "set @rettime = '%d'", atoi(rettime));
  else
    DOFKT(do_, mysql_query)(&mh, "select wert + 0 into @rettime from systemvars "
                                 "where name = 'StoretimeServerdocu'");
  if( do_mysql_query(&mh,
        "insert into serverdoku set serverID=@sr, name='%S', "
          "dateiname='%S', mime='%s', charset='%S', datum=@dt, beschreibung='%S', "
          "deldate = @dt + interval @rettime day, daten='%S'",
        name, q, get_ct(q), charset, beschreibung, wert) )
  { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
    ret = ret | ERROR_SERVERDOKU;
   }

  return ret;
}


int mysql_set_server_suspension(void)
{ long long srvid;

  if( *p_datetime && *p_stopdatetime )
  { if( *ssl )
      srvid = mysql_query_int(&mh, "select @srvid:=getserveridx('%S', '%S')",
                                 p_server, ssl);
    else
      srvid = mysql_query_int(&mh, "select @srvid:=getserveridi('%S', '%S')",
                                 p_server, ip);

    if( ! srvid )
      return ERROR_NO_SERVER;
#ifdef DEBUG
    fprintf(stderr, "set_server_suspension, srvid: %lld.\n", srvid);
#endif

    if( do_mysql_query(&mh, "set @sd = '%S'", p_datetime) )
      return ERROR_QUERY;
    if( *p_stopdatetime == '+' )
    { int i;
      char c;

      if( 2 == sscanf(p_stopdatetime+1, "%d%c", &i, &c) && i > 0 )
      { switch(c)
        { case 'm': do_mysql_query(&mh, "set @ed = @sd + interval %d minute", i);
                    break;
          case 'h': do_mysql_query(&mh, "set @ed = @sd + interval %d hour", i);
                    break;
          case 'd': do_mysql_query(&mh, "set @ed = @sd + interval %d day", i);
                    break;
          default: return ERROR_WERT;
	      }
	    }
    }
    else
    { if( do_mysql_query(&mh, "set @ed = '%S'", p_stopdatetime) )
        return ERROR_QUERY;
	  }
#ifdef DEBUG
    fprintf(stderr, "set_server_suspension, p_stopdatetime: %s.\n", p_stopdatetime);
#endif
    if( mysql_query_int(&mh, "select @ed <= @sd") )
      return ERROR_NOTHING_TO_SET;
    if( do_mysql_query(&mh, "set @accid = getacc('%S')", p_accessright) )
      return ERROR_QUERY;
    if( do_mysql_query(&mh,
            "call test_clientsuspension(@srvid, @accid, '%S', '%S', '%S', '%S', @sd, @ed)",
            p_section, p_teststep, p_detail, p_einheit) )
      return ERROR_QUERY;
    if( mysql_query_int(&mh, "select @ed <= @sd") )
      return ERROR_NO_RIGHTS;         /* Fehler ERROR_NO_RIGHTS wird ignoriert (Code 0) */

                                          /* @srvname wird im Trigger verwendet als setzender Server */
    if( do_mysql_query(&mh, "select verfahrenID, concat(servername, ' (%S)'), concat('^', servername, '$') "
                            "into @domid, @srvname, @s "
                            "from server "
                            "where server.ID = @srvid", ip) )
      return ERROR_QUERY;
    if( do_mysql_query(&mh, "call genreg('%S', @e, @ev)", p_section) )
      return ERROR_QUERY;
    if( do_mysql_query(&mh, "call genreg('%S', @t, @tv)", p_teststep) )
      return ERROR_QUERY;
    if( do_mysql_query(&mh, "call genreg('%S', @d, @dv)", p_detail) )
      return ERROR_QUERY;
    if( do_mysql_query(&mh, "call genreg('%S', @u, @uv)", p_einheit) )
      return ERROR_QUERY;
    if( do_mysql_query(&mh, "set @rem = '%S'", p_remark) )
      return ERROR_QUERY;

    if( mode[1] == 'D' )
    { if( DOFKT(do_, mysql_query)(&mh, "delete from ueberwachungspausen "
           "where verfahrenID = @domid and serversetID = @srvid and server = @s "
             "and server_v = '' and section = @e and section_v = @ev "
             "and teststep = @t and teststep_v = @tv "
             "and details = @d and details_v = @dv and einheit = @u and einheit_v = @uv "
             "and datumvon = @sd and datumbis = @ed and bemerkung = @rem and accessID = @accid") )
        return ERROR_QUERY;
    }
    else if( mode[1] == 'C' )
    { if( DOFKT(do_, mysql_query)(&mh, "update ueberwachungspausen set "
             "datumbis = @ed "
           "where verfahrenID = @domid and serversetID = @srvid and server = @s "
             "and server_v = '' and section = @e and section_v = @ev "
             "and teststep = @t and teststep_v = @tv "
             "and details = @d and details_v = @dv and einheit = @u and einheit_v = @uv "
             "and datumvon = @sd and bemerkung = @rem and accessID = @accid") )
        return ERROR_QUERY;
    }
    else
    { if( DOFKT(do_, mysql_query)(&mh, "insert into ueberwachungspausen "
            "(verfahrenID, serversetID, server, server_v, section, section_v, "
             "teststep, teststep_v, details, details_v, einheit, einheit_v, "
             "datumvon, datumbis, zeitvon, zeitbis, bemerkung, accessID) "
           "values "
            "(@domid, @srvid, @s, '', @e, @ev, @t, @tv, @d, @dv, @u, @uv, "
             "@sd, @ed, '00:00:00', '23:59:59', @rem, @accid)") )
        return ERROR_QUERY;
    }
  }
  return 0;
}

char get(FILE *ptr_in, char *a, int len, char c)
{ unsigned char *p, *q;
  char s;

  s = '\0';
  q = p = (unsigned char *)a;
  while( *z && (*z != c || s != '\0' ) )
  { if( *z == '\xd' || *z == '\xa' )
    { z++;
      if( *z != *(z-1) && (*z == '\xd' || *z == '\xa') )
        z++;
      if( s )
      { if( p-(unsigned char *)a < len-1 )
          *p++ = '\n';
        if( *z )
          continue;
        if( NULL != fgets(folgezeile, MAX_ZEILENLAENGE, ptr_in) )
        { z = folgezeile;
          continue;
        }
        else
        { *p = '\0';
          return '\0';
        }
      }
      else
        break;
    }
    if( *z == '\\' )
    { z++;
      if( *z )
      { if( p-(unsigned char *)a < len-1 )
          *p++ = *z++;
        else
          z++;
      }
    }
    else if( *z == '"' || *z == '\'' )
    { if( s )
      { if( *z == s )
        { s = '\0';
          z++;
          q = p;
        }
        else
        { if( p-(unsigned char *)a < len-1 )
            *p++ = *z++;
          else
            z++;
        }
      }
      else
        s = *z++;
    }
    else
    { if( p-(unsigned char *)a < len-1 )
        *p++ = *z++;
      else
        z++;
    }
  }
  while( p > q && *(p-1) == ' ' )
    p--;
  *p = '\0';
  if( *z )
    return *z++;
  else
    return '\0';
}

char getdatetime(FILE *ptr_in, char *datetime, int lendatetime, char *tz,
                 int lentz, short *microseconds)
{ unsigned char *p;
  char s;

  *tz = s = '\0';
  /****    Date   ***/
  p = (unsigned char *)datetime;
  while( *z && (*z != ' ' && *z != 'T') )
  { if( *z == '\xd' || *z == '\xa' )
    { z++;
      if( *z != *(z-1) && (*z == '\xd' || *z == '\xa') )
        z++;
      if( s )
      { if( !*z )
        { if( NULL != fgets(folgezeile, MAX_ZEILENLAENGE, ptr_in) )
            z = folgezeile;
          else
          { *p = '\0';
            return '\0';
          }
        }
        break;
      }
      else
      { *p = '\0';
        return *z;
      }
    }
    if( *z == '\\' )
    { z++;
      if( *z )
      { if( p-(unsigned char *)datetime < lendatetime-1 )
          *p++ = *z++;
        else
          z++;
      }
    }
    else if( *z == '"' || *z == '\'' )
    { if( s )
      { if( *z == s )
        { s = '\0';
          z++;
        }
        else
        { if( p-(unsigned char *)datetime < lendatetime-1 )
            *p++ = *z++;
          else
            z++;
        }
      }
      else
        s = *z++;
    }
    else
    { if( p-(unsigned char *)datetime < lendatetime-1 )
        *p++ = *z++;
      else
        z++;
    }
  }
  /****    Time   ***/
  while( *z == ' ' || *z == 'T' )
    z++;
  if( microseconds )
    *microseconds = 0;
  *p++ = ' ';
  while( *z && (*z != ';' && *z != ' ' && *z != '+' && *z != '-') )
  { if( *z == '\xd' || *z == '\xa' )
    { z++;
      if( *z != *(z-1) && (*z == '\xd' || *z == '\xa') )
        z++;
      if( s )
      { if( !*z )
        { if( NULL != fgets(folgezeile, MAX_ZEILENLAENGE, ptr_in) )
            z = folgezeile;
          else
          { *p = '\0';
            return '\0';
          }
        }
        break;
      }
      else
      { *p = '\0';
        return *z;
      }
    }
    if( *z == '\\' )
    { z++;
      if( *z )
      { if( p-(unsigned char *)datetime < lendatetime-1 )
          *p++ = *z++;
        else
          z++;
      }
    }
    else if( *z == '"' || *z == '\'' )
    { if( s )
      { if( *z == s )
        { s = '\0';
          z++;
        }
        else
        { if( p-(unsigned char *)datetime < lendatetime-1 )
            *p++ = *z++;
          else
            z++;
        }
      }
      else
        s = *z++;
    }
    else
    { if( p-(unsigned char *)datetime < lendatetime-1 )
      { if( microseconds && *z == '.' )
          *microseconds = 1;
        *p++ = *z++;
      }
      else
        z++;
    }
  }
  *p = '\0';
  /****    Timezone   ***/
  while( *z == ' ' )
    z++;
  if( *z == ';' )
    return *z++;
  p = (unsigned char *)tz;
  while( *z && (*z != ';' || s != '\0' ) )
  { if( *z == '\xd' || *z == '\xa' )
    { z++;
      if( *z != *(z-1) && (*z == '\xd' || *z == '\xa') )
        z++;
      if( s )
      { if( !*z )
        { if( NULL != fgets(folgezeile, MAX_ZEILENLAENGE, ptr_in) )
          { z = folgezeile;
            continue;
          }
          else
          { *p = '\0';
            return '\0';
          }
        }
        break;
      }
      else
      { *p = '\0';
        return *z;
      }
    }
    if( *z == '\\' )
    { z++;
      if( *z )
      { if( p-(unsigned char *)tz < lentz-1 )
          *p++ = *z++;
        else
          z++;
      }
    }
    else if( *z == '"' || *z == '\'' )
    { if( s )
      { if( *z == s )
        { s = '\0';
          z++;
        }
        else
        { if( p-(unsigned char *)tz < lentz-1 )
            *p++ = *z++;
          else
            z++;
        }
      }
      else
        s = *z++;
    }
    else
    { if( p-(unsigned char *)tz < lentz-1 )
        *p++ = *z++;
      else
        z++;
    }
  }
  *p = '\0';
  if( *z )
    return *z++;
  else
    return '\0';
}


int makeuls(FILE *ptr_in)
{ char c;
  int r, ret;

  ret = 0;

  if( charset_changed )
  { mysql_set_character_set(&mh, def_mysql_charset);
    charset_changed = 0;
  }

  datetime[0] = tz[0] = server[0] = section[0] = teststep[0]
    = detail[0] = wert[0] = einheit[0] = accessright[0] = extra[0] = '\0';

  while( !feof(ptr_in) )
  { if( NULL != fgets(zeile, MAX_ZEILENLAENGE, ptr_in) )
    { z = zeile;
#ifdef DEBUG
      fprintf(stderr, "zeile: %s", zeile);
#endif

      skip_blanks(z);
      if( *z == '#' || *z == ';' || NULL == strchr("ACDEFILNOPRSTUVW", *z) )
        continue;
      
      c = get(ptr_in, mode, MAX_MODE_LEN, ';');
      if( c != ';' )
        continue;

#ifdef DEBUG
      fprintf(stderr, "mode: %s, c: %c, z: %s.\n", mode, c, z);
#endif

      if( *mode == 'P' )
      { p_datetime[0] = p_server[0] = p_section[0] = p_teststep[0]
          = p_detail[0] = p_stopdatetime[0] = p_einheit[0]
          = p_accessright[0] = p_remark[0] = '\0';

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
        { c = getdatetime(ptr_in, p_datetime, MAX_DATETIME_LEN, p_tz, MAX_TZ_LEN, NULL);
#ifdef DEBUG
          fprintf(stderr, "p_datetime: %s, ptz: %s, c: %c.\n", p_datetime, p_tz, c);
#endif
          if( c != ';' )
            continue;
        }
#ifdef DEBUG
        fprintf(stderr, "Pdatetime: %s %s\n", p_datetime, p_tz);
#endif

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_server, MAX_SERVER_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_section, MAX_SECTION_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_teststep, MAX_TESTSTEP_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_detail, MAX_DETAIL_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
        { if( *z == '+' )
            c = get(ptr_in, p_stopdatetime, MAX_WERT_LEN, ';');
          else
            c = getdatetime(ptr_in, p_stopdatetime, MAX_DATETIME_LEN, p_stoptz,
                            MAX_TZ_LEN, NULL);
          if( c != ';' )
            continue;
        }

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_einheit, MAX_EINHEIT_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_accessright, MAX_ACCESS_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, p_remark, MAX_EXTRA_LEN, ';');

  #ifdef DEBUG
        fprintf(stderr, "  ip: %s, ssl: %s, datetime: %s\n",
                ip, ssl, p_datetime);
        fprintf(stderr, "  server: %s, section: %s\n", p_server, p_section);
        fprintf(stderr, "  teststep: %s\n", p_teststep);
        fprintf(stderr, "  detail: %s\n", p_detail);
        fprintf(stderr, "  enddt: %s, einheit: %s\n", p_stopdatetime, p_einheit);
        fprintf(stderr, "  accessright: %s\n\n", p_accessright);
  #endif
        if( (r = mysql_set_server_suspension()) )
        { fprintf(stderr, "ERROR %d in line: %s", r, zeile);
          ret = ret | r;
        }
      }
      else
      { wert[0] = einheit[0] = extra[0] = '\0';

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
        { c = getdatetime(ptr_in, datetime, MAX_DATETIME_LEN, tz,
                          MAX_TZ_LEN, &microsecondflag);
          if( c != ';' )
            continue;
        }

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, server, MAX_SERVER_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, section, MAX_SECTION_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, teststep, MAX_TESTSTEP_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, detail, MAX_DETAIL_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, wert, MAX_WERT_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, einheit, MAX_EINHEIT_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, accessright, MAX_ACCESS_LEN, ';');

        skip_blanks(z);
        if( *z == ';' )
          c = *z++;
        else if( c == ';' )
          c = get(ptr_in, extra, MAX_EXTRA_LEN, ';');

  #ifdef DEBUG
        fprintf(stderr, "  ip: %s, ssl: %s, datetime: %s, tz: %s\n",
                ip, ssl, datetime, tz);
        fprintf(stderr, "  server: %s, section: %s\n", server, section);
        fprintf(stderr, "  teststep: %s\n", teststep);
        fprintf(stderr, "  detail: %s\n", detail);
        fprintf(stderr, "  wert: %s, einheit: %s\n", wert, einheit);
        fprintf(stderr, "  accessright: %s\n\n", accessright);
  #endif

        switch( mode[0] )
        { case 'F': r = store_blob(wert, *einheit ? einheit : wert, extra);
                    break;
          case 'I': r = store_blob(wert, *einheit ? einheit : wert, "");
                    break;
          case 'N':
          case 'U':
          case 'O':
          case 'V':
          case 'E':
          case 'C': r = store_value();
                    break;
          case 'R': r = store_section_remark();
                    break;
          case 'T': r = store_teststep_remark();
                    break;
          case 'A': r = store_detail_remark();
                    break;
          case 'S': r = store_server_doku(section, teststep, detail, wert,
                                          *einheit ? einheit : wert, extra);
                    break;
          case 'W': r = store_server_doku_value(section, teststep, detail, wert,
                                                einheit, extra);
                    break;
          case 'L': mysql_set_character_set(&mh, wert);
                    charset_changed = 1;
                    r = 0;
                    break;
          default:  r = 0;
        }
        if( r )
        { if( !(ret & r) || r <= MAX_ERROR_SHOW_EVERY )
          { fprintf(stderr, "ERROR %d from %s (%s) in line: %s", r, server, ip, zeile);
            ret = ret | r;
          }
        }
      }
    }
    else
      return ret;
  }
  return ret;
}


#ifdef SCANDIR
/****************************************************************************************/
/* int select_uls(const struct dirent *d)                                               */
/*                const struct dirent *d: Dirstruktur von scandir                       */
/*     select_uls liefert 1, wenn *d *.uls entspricht                                   */
/****************************************************************************************/
int select_uls(const struct dirent *d)
{
  return !strcasecmp(d->d_name+strlen(d->d_name)-4, ".uls");
}

int makeulsdir(char *indir)
{ FILE *ptr_in;
  struct dirent **namelist;
  int i, n, ret = 0;

  if( (indir[0] != '.' || indir[1] != '\0') && chdir(indir) )
  { fprintf(stderr, "Error on changing directory to %s.\n", indir);
    return 1;
  }

  if( 0 < (n = scandir(".", &namelist, select_uls, alphasort)) )
  { for( i = 0; i < n; i++ )
    { if( NULL != (ptr_in = fopen(namelist[i]->d_name, "r")) )
      { ret = ret | makeuls(ptr_in);
        fclose(ptr_in);
#ifdef DEBUG
        fprintf(stderr, "Done file %s.\n", namelist[i]->d_name);
#endif
      }
      else
        fprintf(stderr, "Error on opening file %s.\n", namelist[i]->d_name);
    }
    free(namelist);
  }
  return ret;
}

#else
int makeulsdir(char *indir)
{ FILE *ptr_in;
  DIR *dirp;
  struct dirent *direntp;
  int ret = 0;

  if( (indir[0] != '.' || indir[1] != '\0') && chdir(indir) )
  { fprintf(stderr, "Error on changing directory to %s.\n", indir);
    return 1;
  }

  if( (dirp = opendir(".")) )
  { while( (direntp = readdir(dirp)) != NULL )
    { if( !strcasecmp(direntp->d_name+strlen(direntp->d_name)-4, ".uls") )
      { if( NULL != (ptr_in = fopen(direntp->d_name, "r")) )
        { ret = ret | makeuls(ptr_in);
          fclose(ptr_in);
#ifdef DEBUG
          fprintf(stderr, "Done file %s.\n", direntp->d_name);
#endif
        }
        else
          fprintf(stderr, "Error on opening file %s.\n", direntp->d_name);
      }
    }
    closedir(dirp);
  }
  return ret;
}
#endif


int main(int argc, char **argv)
{ char inputfile[256];
  FILE *ptr_in;
  char *mysql_pwd = ULS_MYSQL_PWD;
  char *mysql_user = ULS_MYSQL_USER;
  char *mysql_db = ULS_MYSQL_DB;
  int mysql_port = ULS_MYSQL_PORT;
  char *mysql_host = ULS_MYSQL_HOST;
  char *mysql_socket = NULL;
  int options = 0;
  int ret;
  char *indir = NULL;

  lastdatetime[0] = lasttz[0] = lastserver[0] = lastsection[0] = lastteststep[0] = '\0';
 
  while( argc > options+1
         && argv[options+1][0] == '-' && argv[options+1][1] != '\0' )
  { options++;
    switch( argv[options][1] )
    { case 'u': mysql_user = argv[++options];
                break;
      case 'p': mysql_pwd = argv[++options];
                break;
      case 'd': mysql_db = argv[++options];
                break;
      case 'i': ip = argv[++options];
                break;
      case 's': ssl = argv[++options];
                break;
      case 'S': mysql_socket = argv[++options];
                break;
      case 'h': mysql_host = argv[++options];
                break;
      case 'P': mysql_port = atoi(argv[++options]);
                break;
      case 'D': indir = argv[++options];
                break;
      case 'c': def_mysql_charset = argv[++options];
                break;
      case 'v': show_version(true);
                exit(0);
      default: fprintf(stderr, "usage: %s [-S socket|-h host] [-P port] [-u dbuser] [-p dbpwd] [-d dbdatabase] [-c charset] [-i ip|-s ssl] [-D indir] [-v] [-?] [ulsdat ...]\n", argv[0]);
               exit(0);
    }
  }

  mysql_init(&mh);

  if( mysql_socket )
  { if( NULL == mysql_real_connect(&mh, NULL, mysql_user, mysql_pwd,
                                   mysql_db, 0, mysql_socket, 0) )
    { fprintf(stderr, "Failed to connect to database: Error: %s.\n", mysql_error(&mh));
      exit(ERROR_MYSQL_CONNECT);
    }
  }
  else
  { if( NULL == mysql_real_connect(&mh, mysql_host, mysql_user, mysql_pwd,
                                   mysql_db, mysql_port, NULL, 0) )
    { fprintf(stderr, "Failed to connect to database: Error: %s.\n", mysql_error(&mh));
      exit(ERROR_MYSQL_CONNECT);
    }
  }

#ifdef DEBUGSIG
  signal(SIGSEGV, sigsegv_handler);
#endif

  mysql_query(&mh, "set sql_mode = ''");
  mysql_set_character_set(&mh, def_mysql_charset);

  if( indir )
    ret = makeulsdir(indir);
  else
    ret = 0;

  if( argc > 1+options )
  { while( argc > 1+options )
    { if( NULL != (ptr_in = fopen(argv[1+options], "r")) )
      { ret = ret | makeuls(ptr_in);
        fclose(ptr_in);
      }
      else
        fprintf(stderr, "Error on opening inputfile %s.\n",
                argv[1+options]);
      options++;
    }
  }
  else if( !indir )
  { if( isatty(fileno(stdin)) )
    { fprintf(stderr, "inputfile? ");
      scanf("%s", inputfile);
      ptr_in = fopen(inputfile, "r");
    }
    else if( !isatty(fileno(stdin))
             || (inputfile[0] == '-' && inputfile[1] == '\0') )
      ptr_in = stdin;
    else
    { fprintf(stderr, "ERROR: No input!\n");
      exit(ERROR_NO_INPUTFILE);
    }
    if( ptr_in == NULL )
      fprintf(stderr, "Error on opening file %s.\n", inputfile);
    else
    { ret = ret | makeuls(ptr_in);

      if( ptr_in != stdin )
        fclose(ptr_in);
    }
  }
  mysql_close(&mh);
  return ret;
}
