#define MY_PROGRAMMNAME "fileunpack"
#define MY_VERSION "1.3"
#define MY_DATE "16.10.20"
#define MY_COPYRIGHT "H. Sander, Flintbek 2020\n"

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <dirent.h>

#ifdef _LARGEFILE64_SOURCE
#define ULONG unsigned long long
#define ULONGMUSTER "%llu"
#else
#define open64(a,b,c) open(a,b,c)
#define ULONG unsigned long
#define ULONGMUSTER "%lu"
#endif


#define BUF_LEN 10240000
#define MAX_DIR_LEN 4096

#define false 0
#define true !false

int sh;
unsigned int maxfilesize = 0;

char buf[BUF_LEN];

/***************************************************************************************/
/* void usage(char *prg)                                                               */
/*            char *prg: Name, mit dem Programm aufgerufen wurde                       */
/*      usage zeigt Aufrufparameter                                                    */
/***************************************************************************************/
void usage(char *prg)
{ fprintf(stderr, "usage: %7s [-m <maxfilesize>] [packfile]\n", prg);
}


long readdata(int hd_in, char *buffer, long nb)
{ ssize_t rnb, nnb;

  rnb = 0;
  while( rnb < nb )
  { if( 0 < (nnb = read(hd_in, buffer+rnb, nb-rnb)) )
      rnb += nnb;
    if( nnb < 0 )
    { perror("read");
      break;
    }
    else if( nnb == 0 )
      break;
  }

  return rnb;
}

#ifndef OPEN64
#define open64(a,b,c) open(a,b,c)
#endif

int fileread(int hd_in)
{ int hd_out;
  unsigned long nb, rnb, bnb, wnb, wwnb;
  unsigned char len[10];
  char path[MAX_DIR_LEN];
  char *p, *q;

  if( 4 != readdata(hd_in, len, 4) )
  { fprintf(stderr, "Lesefehler!\n");
    return 1;
  }
  nb = len[0] | (len[1]<<8);
  while( nb )
  { if( nb > MAX_DIR_LEN )
    { fprintf(stderr, "Name zu lang!\n");
      return 1;
    }
    if( nb != readdata(hd_in, path, nb) )
    { fprintf(stderr, "Lesefehler!\n");
      return 1;
    }
    path[nb] = '\0';
    if( NULL == (p = strrchr(path, '/')) )
      p = path;
    else
      p++;
    if( NULL != (q = strrchr(p, '\\')) )
      p = q+1;
    if( !*p )
    { fprintf(stderr, "Ungltiger Dateiname!\n");
      return 1;
    }
    if( 0 < (hd_out = open64(p, O_WRONLY | O_CREAT | O_TRUNC,
                            S_IRUSR | S_IWUSR)) )
    { if( 4 != readdata(hd_in, len, 4) )
      { fprintf(stderr, "Lesefehler!\n");
        break;
      }
      nb = len[0] | (len[1]<<8) | (len[2]<<16) | (len[3]<<24);
      while( nb )
      { rnb = 0;
        wnb = 0;
        while( rnb < nb )
        { if( nb-rnb > BUF_LEN )
            bnb = BUF_LEN;
          else
            bnb = nb-rnb;
          if( bnb != readdata(hd_in, buf, bnb) )
          { fprintf(stderr, "Lesefehler!\n");
            return 2;
          }
          if( maxfilesize == 0 )
            wwnb = bnb;
          else if( wnb >= maxfilesize )
            wwnb = 0;
          else if( wnb + bnb > maxfilesize )
            wwnb = maxfilesize - wnb;
          else
            wwnb = bnb;
          if( wwnb > 0 )
          { if( wwnb != write(hd_out, buf, wwnb) )
            { fprintf(stderr, "Schreibfehler!\n");
              return 3;
            }
          }
          rnb += bnb;
          wnb += wwnb;
        }
        if( 4 != readdata(hd_in, len, 4) )
        { fprintf(stderr, "Lesefehler!\n");
          return 2;
        }
        nb = len[0] | (len[1]<<8) | (len[2]<<16) | (len[3]<<24);
      }
      close(hd_out);
    }
    if( 4 != readdata(hd_in, len, 4) )
    { fprintf(stderr, "Lesefehler!\n");
      return 1;
    }
    nb = len[0] | (len[1]<<8);
  }
  return 0;
}


int fileunpack(char *inputfile)
{ int hd_in;
  int ret;
  char packheader[4];

  if( inputfile && *inputfile )
  { if( 0 > (hd_in = open(inputfile, O_RDONLY)) )
    { fprintf(stderr, "Eingabedatei %s konnte nicht geffnet werden.\n", inputfile);
      return 1;
    }
  }
  else
    hd_in = fileno(stdin);

  if( 4 != readdata(hd_in, packheader, 4) )
  { fprintf(stderr, "Lesefehler!\n");
    return 1;
  }

  if(!memcmp(packheader, "FP10", 4) )
    ret = fileread(hd_in);
  else
  { fprintf(stderr, "Falsches Inputformat!\n");
    ret = 2;
  }

  if( inputfile && *inputfile )
  { close(hd_in);
  }
 return ret;
}


int main(int argc, char **argv)
{ int options = 0;
  int stdout_flag = false;

  while( argc > options+1
         && argv[options+1][0] == '-' && argv[options+1][1] != '\0' )
  { options++;
    switch( argv[options][1] )
    { case 'v': fprintf(stderr, MY_PROGRAMMNAME " Version " MY_VERSION ", " MY_DATE ".\n");
                fprintf(stderr, MY_COPYRIGHT);
                exit(0);
      case 'm': if( (maxfilesize = atoi(argv[++options])) <= 0 )
                  maxfilesize = 0;
                break;
      default : usage(argv[0]);
                exit(1);
    }
  }

  if( argc <= 1+options )
    return fileunpack(NULL);
  else
    return fileunpack(argv[options+1]);
}
