/****************************************/
/* mysql_sql.c                          */
/*                                      */
/* Programm zum Lesen von der ULS-DB    */
/****************************************/

#include "mysql_sql.h"


void print_usage(char *prg)
{ fprintf(stderr,
          "usage: %s [-P port] [-u user] [-p pwd] [-d db] [-s server]\n"
          "                [-n nlc] [-c sc] [-Q quotestr] \"sql\" ...\n",
          prg);
}

void subst(char *out, char *in, unsigned long n, char *nl, char *quote)
{ char *s;
  char *p;

  s = in;
  while( s-in <= n )
  { if( nl && (*s == CR || *s == LF) )
    { if( *(s+1) != *s && ( *(s+1) == CR || *(s+1) == LF) )
        s++;
      s++;
      p = nl;
      while( *p )
        *out++ = *p++;
    }
    else if( quote && *s && strchr(quote+1, *s) )
    { *out++ = *quote;
      *out++ = *s++;
    }  
    else
      *out++ = *s++;
  }
  *out = '\0';
}


int main(int argc, char **argv)
{ MYSQL mh;
  MYSQL_RES *mres;
  unsigned int nf, i, binflag;
  MYSQL_ROW mrow;
  char zeile[1024000], query[MAX_QUERYLEN], *q, *a;
  int options = 0;
  char *sqlserver;
  char *sqluser;
  char *sqlpwd;
  char *sqldb;
  char *newline;
  char *trenn, tr[4];
  unsigned int t;
  char *quote_string;
  int sqlport;
  int ret;

  binflag = false;

  sqlserver = SQLSERVER;
  sqluser   = SQLUSER;
  sqlpwd    = SQLPWD;
  sqldb     = SQLDB;
  sqlport   = SQLPORT;
  quote_string = newline = NULL;
  trenn = " ";

  ret = 1;

  while( argc > options+1
         && argv[options+1][0] == '-' && argv[options+1][1] != '\0' )
  { options++;
    switch( argv[options][1] )
    { case 'P': options++;
                sscanf(argv[options], "%d", &sqlport);
                break;
      case 'n': options++;
                newline = argv[options];
                break;
      case 'b': binflag = true;
                break;
      case 'c': options++;
                if( argv[options][0] == '\\' )
                { switch(argv[options][1])
                  { case 'n': trenn = "\n";
                              break;
                    case 't': trenn = "\t";
                              break;
                    case 'x': sscanf(argv[options]+2, "%x", &t);
                              tr[0] = t;
                              tr[1] = '\0';
                              trenn = tr;
                              break;
                    case '\\': trenn = "\\";
                              break;
                    default: if( isdigit(argv[options][1]) )
                             { sscanf(argv[options]+1, "%o", &t);
                               tr[0] = t;
                               tr[1] = '\0';
                               trenn = tr;
                             }
                             else
                               trenn = argv[options];
                  }
                }
                else
                  trenn = argv[options];
                break;
      case 'u': options++;
                sqluser = argv[options];
                break;
      case 'p': options++;
                sqlpwd = argv[options];
                break;
      case 's': options++;
                sqlserver = argv[options];
                break;
      case 'd': options++;
                sqldb = argv[options];
                break;
      case 'Q': options++;
                quote_string = argv[options];
                break;
      case 'v': fprintf(stderr, "%s Version %s, %s.\n", PROGRAMMNAME, VERSION, DATE);
                fprintf(stderr, COPYRIGHT);
                break;
    }
  }

  if( argc <= options + 1 )
  { print_usage(argv[0]);
    exit(1);
  }

  mysql_init(&mh);

  if( NULL != mysql_real_connect(&mh, sqlserver, sqluser, sqlpwd,
                                 sqldb, sqlport, NULL, 0) )

  { mysql_set_character_set(&mh, CHARACTERSET);
    mysql_query(&mh, "set sql_mode = ''");
    for( i = options+1; i < argc-1; i++ )
    { if( mysql_query(&mh, argv[i]) )
      { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
        fprintf(stderr, "sql: %s.\n", argv[i] );
      } 
    }
    a = argv[argc-1];
    while( *a )
    { char c;

      q = query;
      c = '\0';
      while( *a && (*a != ';' || c) && q-query < MAX_QUERYLEN )
      { if( c )
        { if( c == *a )
            c = '\0';
        }
        else if( *a == '\'' || *a == '"' )
          c = *a;
        *q++ = *a++;
      }
      *q = '\0';
      if( *a == ';' )
      { if( mysql_query(&mh, query ) )
        { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
          fprintf(stderr, "sql: %s.\n", query );
        }
        a++;
      }
      else
      { if( mysql_query(&mh, query ) )
        { fprintf(stderr, "mysql_query error: %s.\n", mysql_error(&mh));
          fprintf(stderr, "sql: %s.\n", query );
        }
        else
        { mres = mysql_store_result(&mh);
          if( mres )
          { nf = mysql_num_fields(mres);

            while( (mrow = mysql_fetch_row(mres)) )
            { unsigned long *lengths;

              lengths = mysql_fetch_lengths(mres);
              if( binflag )
              { for( i = 0; i < nf; i++ )
                  fwrite(mrow[i], 1, lengths[i], stdout);
              }
              else
              { if( newline || quote_string )
                { for( i = 0; i < nf; i++ )
                  { if( mrow[i] )
                    { subst(zeile, mrow[i], lengths[i], newline, quote_string);
                      printf("%s%s", i ? trenn : "", zeile);
                    }
                    else
                      printf("%s%.*s", i ? trenn : "", (int) lengths[i], "");
                  }
                }
                else
                { for( i = 0; i < nf; i++ )
                    printf("%s%.*s", i ? trenn : "", (int) lengths[i], mrow[i] ? mrow[i] : "NULL");
                }
                printf("\n");
              }
            }
            mysql_free_result(mres);
          }
          else
          { if( mysql_errno(&mh) )
              fprintf(stderr, "Error: %s\n", mysql_error(&mh));
          }
        }
      }
    }
    mysql_close(&mh);
  }
  else
    fprintf(stderr, "Failed to connect to database: Error: %s.\n", mysql_error(&mh));
  return ret;
}
